/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/exceptions.hh"
#include "src/get_log_entry.hh"
#include "src/util.hh"

#include "libfred/db_settings.hh"
#include "libfred/opcontext.hh"
#include "util/db/query_param.hh"
#include "util/log/logger.hh"

namespace Fred {
namespace Logger {

namespace {

bool get_basic_info(LibFred::OperationContext& _ctx,
        LogEntryInfo& _log_entry_info,
        const LogEntryIdent _ident,
        const unsigned long long _service_id)
{
    const Database::Result db_result = _ctx.get_conn().exec_params(
                // clang-format off
                "SELECT  r.id AS request_id, "
                        "r.time_begin AS time_begin, "
                        "r.time_end AS time_end, "
                        "r.source_ip AS source_ip, "
                        "s.name AS service_name, "
                        "rc.result_code AS result_code, "
                        "rc.name AS result_message, "
                        "rt.name AS request_type, "
                        "rd_req.content AS raw_request, "
                        "rd_res.content AS raw_response, "
                        "r.user_name AS username, "
                        "r.user_id AS user_id "
                "FROM request r "
                "JOIN service s ON s.id = r.service_id "
                "JOIN request_type rt ON rt.id = r.request_type_id "
                "LEFT JOIN result_code rc ON rc.id = r.result_code_id "
                "LEFT JOIN request_data rd_req ON rd_req.request_id = r.id AND NOT rd_req.is_response "
                        "AND rd_req.request_monitoring = $4::boolean "
                        "AND rd_req.request_time_begin = $2::timestamp "
                        "AND rd_req.request_service_id = $3::bigint "
                "LEFT JOIN request_data rd_res ON rd_res.request_id = r.id AND rd_res.is_response "
                        "AND rd_res.request_monitoring = $4::boolean "
                        "AND rd_res.request_time_begin = $2::timestamp "
                        "AND rd_res.request_service_id = $3::bigint "
                "WHERE r.is_monitoring = $4::boolean "
                "AND r.time_begin = $2::timestamp "
                "AND r.service_id = $3::bigint "
                "AND r.id = $1::bigint ",
                // clang-format on
            Database::query_param_list(_ident.id)
                                    (_ident.time_begin)
                                    (_service_id)
                                    (_ident.monitoring_status));

    if (db_result.size() != 1)
    {
        return false;
    }

    _log_entry_info.time_begin = boost::posix_time::time_from_string(
            static_cast<std::string>(db_result[0]["time_begin"]));
    _log_entry_info.service_name =
            static_cast<std::string>(db_result[0]["service_name"]);

    if (!db_result[0]["time_end"].isnull())
    {
        _log_entry_info.time_end = boost::posix_time::time_from_string(
              static_cast<std::string>(db_result[0]["time_end"]));
    }
    if (!db_result[0]["source_ip"].isnull())
    {
        _log_entry_info.source_ip = static_cast<std::string>(db_result[0]["source_ip"]);
    }
    if (!db_result[0]["request_type"].isnull())
    {
        _log_entry_info.log_entry_type = static_cast<std::string>(db_result[0]["request_type"]);
    }
    if (!db_result[0]["result_code"].isnull())
    {
        _log_entry_info.result_code = static_cast<unsigned long long>(db_result[0]["result_code"]);
    }
    if (!db_result[0]["result_message"].isnull())
    {
        _log_entry_info.result_name = static_cast<std::string>(db_result[0]["result_message"]);
    }
    if (!db_result[0]["raw_request"].isnull())
    {
        _log_entry_info.raw_request = static_cast<std::string>(db_result[0]["raw_request"]);
    }
    if (!db_result[0]["raw_response"].isnull())
    {
        _log_entry_info.raw_response = static_cast<std::string>(db_result[0]["raw_response"]);
    }
    if (!db_result[0]["username"].isnull())
    {
        _log_entry_info.username = static_cast<std::string>(db_result[0]["username"]);
    }
    if (!db_result[0]["user_id"].isnull())
    {
        _log_entry_info.user_id = static_cast<unsigned long long>(db_result[0]["user_id"]);
    }

    return true;
}

void get_properties(LibFred::OperationContext& _ctx,
        LogEntryInfo& _log_entry_info,
        const LogEntryIdent _ident,
        const unsigned long long _service_id)
{
    const Database::Result db_result = _ctx.get_conn().exec_params(
                // clang-format off
                "SELECT rpv.id AS property_id, "
                    "rpn.name AS property_type, "
                    "rpv.value AS property_value, "
                    "child_rpn.name AS child_type, "
                    "child.value AS child_value "
                "FROM request_property_value rpv "
                "JOIN request_property_name rpn ON rpn.id=rpv.property_name_id "
                "LEFT JOIN request_property_value child ON child.parent_id = rpv.id "
                    "AND child.request_monitoring = $4::boolean "
                    "AND child.request_service_id = $3::bigint "
                    "AND child.request_time_begin = $2::timestamp "
                    "AND child.request_id = $1::bigint "
                "LEFT JOIN request_property_name child_rpn ON child_rpn.id = child.property_name_id "
                "WHERE rpv.request_monitoring = $4::boolean "
                    "AND rpv.request_service_id = $3::bigint "
                    "AND rpv.request_time_begin = $2::timestamp "
                    "AND rpv.request_id = $1::bigint "
                    "AND rpv.parent_id IS NULL "
                "ORDER BY rpn.name, rpv.id ",
                // clang-format on
            Database::query_param_list(_ident.id)
                                    (_ident.time_begin)
                                    (_service_id)
                                    (_ident.monitoring_status));

    std::string previous_property_type;
    unsigned long long previous_property_id = 0;
    LogEntryProperties props;
    LogEntryProperty prop;
    for (Database::Result::size_type i = 0; i < db_result.size(); ++i)
    {
        const auto property_id = static_cast<unsigned long long>(db_result[i]["property_id"]);
        const auto property_type = static_cast<std::string>(db_result[i]["property_type"]);
        const auto property_value = static_cast<std::string>(db_result[i]["property_value"]);

        const bool new_property_id = property_id != previous_property_id;
        if (new_property_id)
        {
            if (!prop.property_value.empty() && prop.children.empty())
            {
                props.properties.push_back(prop);
            }
            else if (!prop.children.empty())
            {
                props.properties.push_back(prop);
                prop.children.clear();
            }
            prop.property_value = property_value;

            const bool new_property_type = property_type != previous_property_type;
            if (new_property_type)
            {
                if (!props.properties.empty())
                {
                    _log_entry_info.log_entry_properties.push_back(props);
                    props.properties.clear();
                }
                props.property_type = property_type;
            }
        }

        const bool existing_child = !db_result[i]["child_value"].isnull();
        if (existing_child)
        {
            LogEntryPropertyChildren child;
            child.property_type = static_cast<std::string>(db_result[i]["child_type"]);
            child.property_value = static_cast<std::string>(db_result[i]["child_value"]);
            prop.children.push_back(child);
        }

        previous_property_type = property_type;
        previous_property_id = property_id;
    }
    props.properties.push_back(prop);
    _log_entry_info.log_entry_properties.push_back(props);
}

void get_object_references(LibFred::OperationContext& _ctx,
        LogEntryInfo& _log_entry_info,
        const LogEntryIdent _ident,
        const unsigned long long _service_id)
{
    const Database::Result db_result = _ctx.get_conn().exec_params(
                // clang-format off
                "SELECT rot.name AS object_type_name, "
                    "ror.object_id AS object_id "
                "FROM request_object_ref ror "
                "LEFT JOIN request_object_type rot ON rot.id=ror.object_type_id "
                "WHERE ror.request_monitoring = $4::boolean "
                    "AND ror.request_service_id = $3::bigint "
                    "AND ror.request_time_begin = $2::timestamp "
                    "AND ror.request_id = $1::bigint "
                "ORDER BY rot.name",
                // clang-format on
            Database::query_param_list(_ident.id)
                                    (_ident.time_begin)
                                    (_service_id)
                                    (_ident.monitoring_status));

    std::string previous_type;
    ObjectReferences refs;
    for (Database::Result::size_type i = 0; i < db_result.size(); ++i)
    {
        const auto object_id = static_cast<std::string>(db_result[i]["object_id"]);
        const auto object_type = static_cast<std::string>(db_result[i]["object_type_name"]);

        if (object_type == previous_type)
        {
            refs.object_reference_ids.push_back(object_id);
        }
        else
        {
            if (!refs.object_reference_ids.empty())
            {
                _log_entry_info.object_references.push_back(refs);
                refs.object_reference_ids.clear();
            }
            refs.object_type = object_type;
            refs.object_reference_ids.push_back(object_id);
        }
        previous_type = object_type;
    }
    _log_entry_info.object_references.push_back(refs);
}

} // namespace Fred::Logger::{anonymous}

GetLogEntryInfoReply get_log_entry(const GetLogEntryInfoRequest& _request)
{
    LibFred::OperationContextCreator ctx;
    const unsigned long long service_id = get_service_id_by_name(ctx, _request.log_entry_ident.service_handle);

    LogEntryInfo log_entry_info;

    const bool log_entry_exist = get_basic_info(ctx, log_entry_info, _request.log_entry_ident, service_id);
    if (!log_entry_exist)
    {
        throw LogEntryDoesNotExist();
    }

    get_properties(ctx, log_entry_info, _request.log_entry_ident, service_id);

    get_object_references(ctx, log_entry_info, _request.log_entry_ident, service_id);

    GetLogEntryInfoReply reply;
    reply.log_entry_info = log_entry_info;
    return reply;
}

} // namespace Fred::Logger
} // namespace Fred
