/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/exceptions.hh"
#include "src/list_log_entries.hh"
#include "src/util.hh"

#include "libfred/db_settings.hh"
#include "libfred/opcontext.hh"
#include "util/db/query_param.hh"
#include "util/log/logger.hh"

#include <sstream>

namespace Fred {
namespace Logger {

ListLogEntriesReply list_log_entry_ids(const ListLogEntriesRequest& _request)
{
    if (_request.from.is_special())
    {
        throw InvalidTimestampFrom();
    }
    if (_request.to.is_special())
    {
        throw InvalidTimestampTo();
    }
    else if (_request.to <= _request.from)
    {
        throw InvalidTimestampRange();
    }
    if (_request.count_limit < 1)
    {
        throw InvalidCountLimit();
    }

    LibFred::OperationContextCreator ctx;
    const unsigned long long service_id = get_service_id_by_name(ctx, _request.service);

    Database::QueryParams params;
    std::ostringstream sql_query;

    sql_query << "SELECT r.id AS request_id, r.time_begin AS time_begin, r.service_id AS service_id "
                 "FROM request r "
                 "JOIN request_type rt ON rt.id = r.request_type_id ";

    const bool has_object_reference = (_request.object_reference != boost::none &&
            !_request.object_reference->object_type.empty() &&
            !_request.object_reference->object_id.empty());

    if (has_object_reference)
    {
        sql_query << "LEFT JOIN request_object_ref ror ON ror.request_id = r.id "
                     "AND NOT ror.request_monitoring ";

        params.push_back(service_id);
        sql_query << "AND ror.request_service_id = $" << params.size() << "::bigint ";

        params.push_back(_request.from);
        sql_query << "AND ror.request_time_begin >= $" << params.size() << "::timestamp ";

        params.push_back(_request.to);
        sql_query << "AND ror.request_time_begin < $" << params.size() << "::timestamp "
                     "LEFT JOIN request_object_type rot ON rot.id = ror.object_type_id ";
    }
    sql_query << "WHERE NOT r.is_monitoring ";

    params.push_back(_request.from);
    sql_query << "AND r.time_begin >= $" << params.size() << "::timestamp ";

    params.push_back(_request.to);
    sql_query << "AND r.time_begin < $" << params.size() << "::timestamp ";

    params.push_back(service_id);
    sql_query << "AND r.service_id = $" << params.size() << "::bigint ";

    if (_request.username != boost::none && !_request.username->empty())
    {
        params.push_back(*_request.username);
        sql_query << "AND r.user_name = $" << params.size() << "::text ";
    }

    const unsigned long long empty_user_id = 0;
    if (_request.user_id != boost::none && *_request.user_id != empty_user_id)
    {
        params.push_back(*_request.user_id);
        sql_query << "AND r.user_id = $" << params.size() << "::bigint ";
    }

    std::ostringstream sql_log_entry_types;
    sql_log_entry_types << "AND rt.name IN (";
    std::string delimiter;
    bool has_one_type_at_least = false;
    for (const auto& type : _request.log_entry_types)
    {
        if (!type.empty())
        {
            params.push_back(type);
            sql_log_entry_types << delimiter << "$" << params.size() << "::text";
            if (delimiter.empty())
            {
                delimiter = ", ";
            }
            has_one_type_at_least = true;
        }
    }
    sql_log_entry_types << ") ";
    if (has_one_type_at_least)
    {
        sql_query << sql_log_entry_types.str();
    }

    if (has_object_reference)
    {
        params.push_back((*_request.object_reference).object_id);
        sql_query << "AND ror.object_id = $" << params.size() << "::bigint ";
        params.push_back((*_request.object_reference).object_type);
        sql_query << "AND rot.name = $" << params.size() << "::text ";
    }

    params.push_back(_request.count_limit);
    sql_query << "ORDER BY r.time_begin DESC LIMIT $" << params.size() << "::integer ";

    ListLogEntriesReply reply;
    try
    {
       const Database::Result db_result = ctx.get_conn().exec_params(sql_query.str(), params);
       for (Database::Result::size_type i = 0; i < db_result.size(); ++i)
       {
           LogEntryIdent ident;
           ident.id = static_cast<unsigned long long>(db_result[i]["request_id"]);
           ident.time_begin = boost::posix_time::time_from_string(
                   static_cast<std::string>(db_result[i]["time_begin"]));
           ident.service_handle = _request.service;
           ident.monitoring_status = false;
           reply.log_entry_ids.push_back(ident);
       }
    }
    catch (const std::exception& e)
    {
        LOGGER.warning(e.what());
        throw ListLogEntriesException();
    }
    return reply;
}

} // namespace Fred::Logger
} // namespace Fred
