/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/cfg.hh"
#include "src/exceptions.hh"
#include "src/get_log_entry.hh"
#include "src/list_log_entries.hh"
#include "src/proto_unwrap.hh"
#include "src/proto_wrap.hh"

#include "fred_api/logger/service_search_logger_history_grpc.grpc.pb.h"
#include "fred_api/logger/service_search_logger_history_grpc.pb.h"

#include "libfred/db_settings.hh"
#include "util/log/logger.hh"
#include "util/log/add_log_device.hh"

#include <grpc++/grpc++.h>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/optional.hpp>

#include <csignal>
#include <iostream>
#include <string>

namespace Fred {
namespace Logger {

namespace {
class SearchLoggerHistoryService final : public Api::SearchLoggerHistory::Service
{
    grpc::Status list_log_entries(
            grpc::ServerContext*,
            const Api::ListLogEntriesRequest* _request,
            Api::ListLogEntriesReply* _response) override
    {
        LOGGER.info("server: list_log_entries requested");
        ListLogEntriesRequest request = proto_unwrap(*_request);
        try {
            const ListLogEntriesReply result = list_log_entry_ids(request);
            proto_wrap(result, _response);
            return grpc::Status::OK;
        }
        catch (const ServiceNotFound& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _response);
            return grpc::Status::OK;
        }
        catch (const InvalidTimestampFrom& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _response);
            return grpc::Status::OK;
        }
        catch (const InvalidTimestampTo& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _response);
            return grpc::Status::OK;
        }
        catch (const InvalidTimestampRange& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _response);
            return grpc::Status::OK;
        }
        catch (const InvalidCountLimit& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _response);
            return grpc::Status::OK;
        }
        catch (const ListLogEntriesException& e)
        {
            LOGGER.warning(std::string("server: ") + e.what());
            return grpc::Status(grpc::StatusCode::UNKNOWN, e.what());
        }
        catch (const std::exception& e)
        {
            LOGGER.error(std::string("server caught unexpected exception: ") + e.what());
            return grpc::Status(grpc::StatusCode::UNKNOWN, std::string("unexpected exception: ") + e.what());
        }
        catch (...)
        {
            LOGGER.error("server caught unexpected exception");
            return grpc::Status(grpc::StatusCode::UNKNOWN, "unexpected exception");
        }
    }

    grpc::Status get_log_entry_info(
            grpc::ServerContext*,
            const Api::GetLogEntryInfoRequest* _request,
            Api::GetLogEntryInfoReply* _response) override
    {
        LOGGER.info("server: get_log_entry_info requested");
        try {
            const GetLogEntryInfoRequest request = proto_unwrap(*_request);
            const GetLogEntryInfoReply result = get_log_entry(request);
            proto_wrap(result, _response);
            return grpc::Status::OK;
        }
        catch (const InvalidLogEntryIdent& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(LogEntryDoesNotExist(), _response);
            return grpc::Status::OK;
        }
        catch (const ServiceNotFound& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(LogEntryDoesNotExist(), _response);
            return grpc::Status::OK;
        }
        catch (const LogEntryDoesNotExist& e)
        {
            LOGGER.info(std::string("server: ") + e.what());
            proto_wrap(e, _response);
            return grpc::Status::OK;
        }
        catch (const std::exception& e)
        {
            LOGGER.error(std::string("server caught unexpected exception: ") + e.what());
            return grpc::Status(grpc::StatusCode::UNKNOWN, std::string("unexpected exception: ") + e.what());
        }
        catch (...)
        {
            LOGGER.error("server caught unexpected exception");
            return grpc::Status(grpc::StatusCode::UNKNOWN, "unexpected exception");
        }
    }
};

void run_server(const std::string& server_address)
{
    ::sigset_t blocked_signals;

    // Other threads created by grpc::Server will inherit a copy of the signal mask.
    ::sigemptyset(&blocked_signals);
    ::sigaddset(&blocked_signals, SIGHUP);
    ::sigaddset(&blocked_signals, SIGINT);
    ::sigaddset(&blocked_signals, SIGQUIT);
    ::sigaddset(&blocked_signals, SIGTERM);
    const auto pthread_sigmask_result = ::pthread_sigmask(SIG_BLOCK, &blocked_signals, nullptr);
    constexpr int operation_success = 0;
    if (pthread_sigmask_result != operation_success)
    {
        throw std::runtime_error(std::strerror(pthread_sigmask_result));
    }

    SearchLoggerHistoryService search_logger_history_service;

    grpc::ServerBuilder builder;

    // Listen on the given address without any authentication mechanism.
    builder.AddListeningPort(server_address, grpc::InsecureServerCredentials());

    builder.RegisterService(&search_logger_history_service);

    // Finally assemble the server.
    std::unique_ptr<grpc::Server> server(builder.BuildAndStart());
    LOGGER.info("server: listen on " + server_address);

    int received_signal;
    const auto sigwait_result = ::sigwait(&blocked_signals, &received_signal);
    if (sigwait_result != operation_success)
    {
        throw std::runtime_error(std::strerror(sigwait_result));
    }
    switch (received_signal)
    {
        case SIGHUP:
            LOGGER.debug("SIGHUP received");
            break;
        case SIGINT:
            LOGGER.debug("SIGINT received");
            break;
        case SIGQUIT:
            LOGGER.debug("SIGQUIT received");
            break;
        case SIGTERM:
            LOGGER.debug("SIGTERM received");
            break;
        default:
            throw std::runtime_error("unexpected signal received");
    }
    const std::chrono::system_clock::time_point deadline = std::chrono::system_clock::now() + std::chrono::seconds(10);
    server->Shutdown(deadline);
}

} //namespace Fred::Logger::{anonymous}
} //namespace Fred::Logger
} //namespace Fred

namespace {

enum class LogHandler
{
    is_set,
    is_not_set
};

class SetLogHandler : public boost::static_visitor<LogHandler>
{
public:
    explicit SetLogHandler(Logging::Log& logger_instance) : logger_instance_(logger_instance) { }
    LogHandler operator()(const boost::blank&)const
    {
        return LogHandler::is_not_set;
    }
    LogHandler operator()(const Cfg::Options::Console& option)const
    {
        Logging::add_console_device(logger_instance_, option.min_severity);
        return LogHandler::is_set;
    }
    LogHandler operator()(const Cfg::Options::LogFile& option)const
    {
        Logging::add_file_device(logger_instance_, option.file_name, option.min_severity);
        return LogHandler::is_set;
    }
    LogHandler operator()(const Cfg::Options::SysLog& option)const
    {
        Logging::add_syslog_device(logger_instance_, option.facility, option.min_severity);
        return LogHandler::is_set;
    }
private:
    Logging::Log& logger_instance_;
};

std::string make_connect_string(const Cfg::Options::Database& option)
{
    std::string result;
    if (option.host != boost::none)
    {
        result = "host=" + *option.host;
    }
    if (option.port != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "port=" + std::to_string(*option.port);
    }
    if (option.user != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "user=" + *option.user;
    }
    if (option.dbname != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "dbname=" + *option.dbname;
    }
    if (option.password != boost::none)
    {
        if (!result.empty())
        {
            result += " ";
        }
        result += "password=" + *option.password;
    }
    return result;
}

}//namespace::{anonymous}

int main(int argc, char** argv)
{
    try
    {
        Cfg::Options::init(argc, argv);
    }
    catch (const Cfg::HelpRequested& help)
    {
        std::cout << help.what() << std::endl;
        return EXIT_SUCCESS;
    }
    catch (const Cfg::UnknownOption& unknown_option)
    {
        std::cerr << unknown_option.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const Cfg::MissingOption& missing_option)
    {
        std::cerr << missing_option.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const Cfg::Exception& e)
    {
        std::cerr << "Configuration problem: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }

    switch (boost::apply_visitor(SetLogHandler(LOGGER), Cfg::Options::get().log))
    {
        case LogHandler::is_set:
            LOGGER.debug("fred-logger-services configuration successfully loaded");
            break;
        case LogHandler::is_not_set:
            std::cout << "fred-logger-services configuration successfully loaded" << std::endl;
            break;
    }
    Database::emplace_default_manager<Database::StandaloneManager>(
            make_connect_string(Cfg::Options::get().database));

    try
    {
        Fred::Logger::run_server(Cfg::Options::get().logger.listen_on);
        LOGGER.info("Server shutdown");
        return EXIT_SUCCESS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }
}
