/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/exceptions.hh"
#include "src/proto_unwrap.hh"

#include "util/log/logger.hh"

#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/optional.hpp>

#include <google/protobuf/util/time_util.h>

#include <string>
#include <vector>

namespace Fred {
namespace Logger {

namespace {

boost::posix_time::ptime proto_unwrap(const google::protobuf::Timestamp& _src)
{
    const boost::posix_time::ptime timestamp(boost::gregorian::date(1970, 1, 1),
           (boost::posix_time::seconds(_src.seconds()) +
           boost::posix_time::microseconds(_src.nanos()/1000)));
    return timestamp;
}

ObjectReference proto_unwrap(const Api::ObjectReference& _src)
{
    ObjectReference object_reference;
    if (_src.has_type())
    {
        object_reference.object_type = _src.type().value();
    }
    if (_src.has_value())
    {
        object_reference.object_id = _src.value().value();
    }
    return object_reference;
}

LogEntryIdent proto_unwrap(const Api::LogEntryId& _src)
{
    std::vector<std::string> result;
    boost::split(result, _src.value(), boost::is_any_of("#"));
    LogEntryIdent ident;
    if (result.size() != 4)
    {
        LOGGER.warning(std::string("Invalid format: log_entry_ident = ") + _src.value());
        throw InvalidLogEntryIdent();
    }

    try
    {
        const auto id = boost::lexical_cast<long long>(result.at(0));
        if (id < 1)
        {
            throw LogEntryIdOutOfRange();
        }
        ident.id = id;
    }
    catch (const LogEntryIdOutOfRange& e)
    {
        LOGGER.warning(e.what() + std::string(" log_entry_ident = ") + _src.value());
        throw InvalidLogEntryIdent();
    }
    catch (const boost::bad_lexical_cast&)
    {
        LOGGER.warning(std::string("Could not parse log_entry_id: log_entry_ident = ") + _src.value());
        throw InvalidLogEntryIdent();
    }

    try
    {
        ident.time_begin = boost::posix_time::from_iso_string(result.at(1));
    }
    catch (const std::exception&)
    {
        LOGGER.warning(std::string("Could not parse time_begin: log_entry_ident = ") + _src.value());
        throw InvalidLogEntryIdent();
    }

    ident.service_handle = result.at(2);
    if (ident.service_handle.empty())
    {
        LOGGER.warning(std::string("Could not parse log_entry_service: log_entry_ident = ") + _src.value());
        throw InvalidLogEntryIdent();
    }

    try
    {
        ident.monitoring_status = boost::lexical_cast<bool>(result.at(3));
    }
    catch (const boost::bad_lexical_cast&)
    {
        LOGGER.info(std::string("Could not parse monitoring_status: log_entry_ident = ") + _src.value());
        throw InvalidLogEntryIdent();
    }

    return ident;
}

} // namespace Fred::Logger::{anonymous}

ListLogEntriesRequest proto_unwrap(const Api::ListLogEntriesRequest& _src)
{
    ListLogEntriesRequest request;

    try
    {
        if (_src.has_username())
        {
            request.username = _src.username().value();
        }
        if (_src.has_user_id())
        {
            request.user_id = _src.user_id().value();
        }
        if (_src.has_timestamp_from())
        {
            request.from = proto_unwrap(_src.timestamp_from());
        }
        if (_src.has_timestamp_to())
        {
            request.to = proto_unwrap(_src.timestamp_to());
        }
        if (_src.has_service())
        {
            request.service = _src.service().value();
        }
        for (const auto& type :_src.log_entry_types())
        {
            request.log_entry_types.push_back(type.value());
        }
        if (_src.has_reference())
        {
            request.object_reference = proto_unwrap(_src.reference());
        }
        request.count_limit = _src.count_limit();
    }
    catch (const std::exception& e)
    {
        LOGGER.info(std::string("Unwrap Api::ListLogEntriesRequest: ") + e.what());
        throw;
    }

    return request;
}

GetLogEntryInfoRequest proto_unwrap(const Api::GetLogEntryInfoRequest& _src)
{
    GetLogEntryInfoRequest request;
    try
    {
        request.log_entry_ident = proto_unwrap(_src.log_entry_id());
    }
    catch (const std::exception& e)
    {
        LOGGER.info(std::string("Unwrap Api::GetLogEntryInfoRequest: ") + e.what());
        throw;
    }
    return request;
}

} // namespace Fred::Logger
} // namespace Fred

