/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/proto_wrap.hh"

#include <boost/date_time/posix_time/posix_time.hpp>

#include <sstream>

namespace Fred {
namespace Logger {

namespace {

void proto_wrap(const boost::posix_time::ptime& _src, google::protobuf::Timestamp* _dst)
{
    const boost::posix_time::ptime epoch(boost::gregorian::date(1970, boost::date_time::Jan, 1),
            boost::posix_time::time_duration(0, 0, 0));
    const boost::posix_time::time_duration diff = _src - epoch;
    long long total_seconds = diff.total_seconds();
    long long fractional_seconds = diff.total_nanoseconds() % 1'000'000'000;
    if (diff.is_negative())
    {
        total_seconds--;
        fractional_seconds += 1'000'000'000;
    }
    _dst->set_seconds(total_seconds);
    _dst->set_nanos(fractional_seconds);
}

void proto_wrap(
        const LogEntryProperties& _src,
        google::protobuf::Map< ::std::string, Api::LogEntryProperties >* _dst)
{
    Api::LogEntryProperties props;
    for (const auto& prop : _src.properties)
    {
        const auto& dst_prop = props.add_values();
        dst_prop->set_value(prop.property_value);
        for (const auto& child : prop.children)
        {
            (*dst_prop->mutable_children())[child.property_type] = child.property_value;
        }
    }
    (*_dst)[_src.property_type] = props;
}

void proto_wrap(
        const ObjectReferences& _src,
        google::protobuf::Map< ::std::string, Api::ObjectReferenceValues >* _dst)
{
    Api::ObjectReferenceValues object_refs;
    for (const auto& id : _src.object_reference_ids)
    {
        object_refs.add_references()->set_value(id);
    }
    (*_dst)[_src.object_type] = object_refs;
}

} // namespace Fred::Logger::{anonymous}

void proto_wrap(const ListLogEntriesReply& _src, Api::ListLogEntriesReply* _dst)
{
    for (const auto& ident : _src.log_entry_ids)
    {
        std::ostringstream s;
        s << ident.id << "#"
                << boost::posix_time::to_iso_string(ident.time_begin) << "#"
                << ident.service_handle << "#"
                << ident.monitoring_status;
        _dst->mutable_data()->add_log_entry_ids()->set_value(s.str());
    }
}

void proto_wrap(const ServiceNotFound&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_log_entry_service_not_found();
}

void proto_wrap(const InvalidTimestampFrom&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_timestamp_from();
}

void proto_wrap(const InvalidTimestampTo&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_timestamp_to();
}

void proto_wrap(const InvalidTimestampRange&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_timestamp_range();
}

void proto_wrap(const InvalidCountLimit&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_count_limit();
}

void proto_wrap(const GetLogEntryInfoReply& _src, Api::GetLogEntryInfoReply* _dst)
{
    Api::LogEntryInfo* dst_info = _dst->mutable_data()->mutable_log_entry_info();
    const LogEntryInfo src_info = _src.log_entry_info;

    proto_wrap(src_info.time_begin, dst_info->mutable_time_begin());

    dst_info->mutable_service()->set_value(src_info.service_name);
    if (!src_info.time_end.is_special())
    {
        proto_wrap(src_info.time_end, dst_info->mutable_time_end());
    }
    if (src_info.source_ip)
    {
        dst_info->mutable_source_ip()->set_value(src_info.source_ip.get());
    }
    if (src_info.log_entry_type)
    {
        dst_info->mutable_log_entry_type()->set_value(src_info.log_entry_type.get());
    }
    if (src_info.result_code)
    {
        dst_info->set_result_code(src_info.result_code.get());
    }
    if (src_info.result_name)
    {
        dst_info->set_result_name(src_info.result_name.get());
    }
    if (src_info.raw_request)
    {
        dst_info->mutable_raw_request()->set_value(src_info.raw_request.get());
    }
    if (src_info.raw_response)
    {
        dst_info->mutable_raw_response()->set_value(src_info.raw_response.get());
    }
    if (src_info.username)
    {
        dst_info->mutable_username()->set_value(src_info.username.get());
    }
    if (src_info.user_id)
    {
        dst_info->mutable_user_id()->set_value(src_info.user_id.get());
    }
    for (const auto& props : src_info.log_entry_properties)
    {
        proto_wrap(props, dst_info->mutable_properties());
    }
    for (const auto& refs : src_info.object_references)
    {
        proto_wrap(refs, dst_info->mutable_references());
    }
}

void proto_wrap(const LogEntryDoesNotExist&, Api::GetLogEntryInfoReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_log_entry_does_not_exist();
}

} // namespace Fred::Logger
} // namespace Fred
