/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/exceptions.hh"
#include "src/list_log_entries.hh"

#include <boost/test/unit_test.hpp>

#include <boost/date_time/posix_time/posix_time.hpp>

#include <iostream>
#include <string>
#include <vector>

namespace Test {

namespace {
constexpr char service_name[] = "MojeID";
const boost::posix_time::ptime to_date = boost::posix_time::second_clock::local_time();
const boost::posix_time::ptime from_date = to_date - boost::gregorian::days(10);
constexpr unsigned int count_limit = 30;
constexpr char username[] = "REG-FRED_A";
constexpr unsigned long long user_id = 1;
constexpr char object_type[] = "contact";
constexpr unsigned long long object_id = 222;
const std::vector<std::string> log_entry_types{
                                        "OpenIDRequest",
                                        "Login",
                                        "Logout",
                                        "UserChange",
                                        "PasswordResetReques"};
}// namespace Test::{anonymous}

BOOST_AUTO_TEST_SUITE(TestListLogEntries)

BOOST_AUTO_TEST_CASE(service_not_found)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = "nonexistent_service";
    request.from = from_date;
    request.to = to_date;
    request.count_limit = count_limit;

    BOOST_CHECK_THROW(::Fred::Logger::list_log_entry_ids(request),
            ::Fred::Logger::ServiceNotFound);
}

BOOST_AUTO_TEST_CASE(bad_from_date)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = service_name;
    request.to = to_date;
    request.count_limit = count_limit;

    BOOST_CHECK_THROW(::Fred::Logger::list_log_entry_ids(request),
            ::Fred::Logger::InvalidTimestampFrom);
}

BOOST_AUTO_TEST_CASE(bad_to_date)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = service_name;
    request.from = from_date;
    request.count_limit = count_limit;

    BOOST_CHECK_THROW(::Fred::Logger::list_log_entry_ids(request),
            ::Fred::Logger::InvalidTimestampTo);
}

BOOST_AUTO_TEST_CASE(bad_date_range)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = service_name;
    request.from = from_date;
    request.to = from_date;
    request.count_limit = count_limit;

    BOOST_CHECK_THROW(::Fred::Logger::list_log_entry_ids(request),
            ::Fred::Logger::InvalidTimestampRange);
}

BOOST_AUTO_TEST_CASE(bad_count_limit)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = service_name;
    request.from = from_date;
    request.to = to_date;
    request.count_limit = 0;

    BOOST_CHECK_THROW(::Fred::Logger::list_log_entry_ids(request),
            ::Fred::Logger::InvalidCountLimit);
}

BOOST_AUTO_TEST_CASE(test_list_log_entries_empty_reference)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = service_name;
    request.from = from_date;
    request.to = to_date;
    request.count_limit = count_limit;
    ::Fred::Logger::ObjectReference object_reference;
    object_reference.object_type = object_type;
    object_reference.object_id = "";
    request.object_reference = object_reference;
    request.log_entry_types = log_entry_types;

    ::Fred::Logger::ListLogEntriesReply reply = ::Fred::Logger::list_log_entry_ids(request);
    BOOST_CHECK(reply.log_entry_ids.empty());
}

BOOST_AUTO_TEST_CASE(test_list_log_entries_min)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = service_name;
    request.from = from_date;
    request.to = to_date;
    request.count_limit = count_limit;

    ::Fred::Logger::ListLogEntriesReply reply = ::Fred::Logger::list_log_entry_ids(request);
    BOOST_CHECK(reply.log_entry_ids.empty());
}

BOOST_AUTO_TEST_CASE(test_list_log_entries_max)
{
    ::Fred::Logger::ListLogEntriesRequest request;
    request.service = service_name;
    request.from = from_date;
    request.to = to_date;
    request.count_limit = count_limit;

    request.username = username;
    request.user_id = user_id;
    ::Fred::Logger::ObjectReference object_reference;
    object_reference.object_type = object_type;
    object_reference.object_id = object_id;
    request.object_reference = object_reference;
    request.log_entry_types = log_entry_types;

    ::Fred::Logger::ListLogEntriesReply reply = ::Fred::Logger::list_log_entry_ids(request);
    BOOST_CHECK(reply.log_entry_ids.empty());
}

BOOST_AUTO_TEST_SUITE_END() // TestListLogEntries

} // namespace Test
